﻿using FlyAd2021.Inc;
using FlyADBase;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.IO.Ports;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Threading;

namespace FlyAd2021
{
    public class FlyAd2021 : IFlyAD
    {
        #region INotifyPropertyChanged 成员
        protected void NotifyPropertyChanged(string propertyName)
        {
            if (this.PropertyChanged != null)
                this.PropertyChanged.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion

        #region IFlyAD
        /// <summary>
        /// AD盒地址, 当为 TCP通讯时，为 192.168.251.200:20020;
        /// 当为 串口通讯时， 为 COM6
        /// </summary>
        public string Addr { get; set; }

        /// <summary>
        /// 从flyad7 获取 的systick 转换的时间
        /// </summary>
        public DateTime Now { get; private set; }

        /// <summary>
        /// AD盒版本号 版本1 没有systick;版本2 有systick;
        /// 通过通讯读取
        /// </summary>
        public int Version { get; } = 2021;

        /// <summary>
        /// 硬件版本，设置的！！！
        /// </summary>
        public int HardwareVersion { get; set; }

        /// <summary>
        /// 脉冲
        /// </summary>
        public int Position { get; private set; }

        /// <summary>
        /// 速度 pps
        /// </summary>
        public int Speed { get; private set; }

        /// <summary>
        /// 纵向脉冲，也叫主轴脉冲
        /// </summary>
        public int Position2 { get; private set; }
        /// <summary>
        /// 纵向速度
        /// </summary>
        public int Speed2 { get; private set; }

        /// <summary>
        /// 1000ms从AD卡获取(或推送)1个AD值。此AD值只用于调试，定点
        /// </summary>
        public int AD { get; private set; }

        /// <summary>
        /// AD值最大值 ，恒定返回65535
        /// </summary>
        public int ADMax => 0xFFFF;

        /// <summary>
        /// 输入口状态,只有12位有效
        /// </summary>
        public UInt16 IStatus { get; private set; }

        /// <summary>
        /// 输出口状态,只有4位有效
        /// </summary>
        public UInt16 OStatus { get; private set; }

        #endregion


        /// <summary>
        /// 只要connect成功，获取systick被复位，都会从AD盒设备读取参数；
        /// 否则， 设置参数 到 AD盒
        /// </summary>
        public bool IsReadParamFromDev { get; set; }



        TSysTick mSysTick = new TSysTick();

        IFlyAd2021Comm comm;
        FlyAd2021Core core;

        public FlyAd2021()
        {
            
        }

        public void Init() 
        {
            if (Addr.ToLower().StartsWith("com"))
            {
                comm = new FlyAd2021Comm_SerialPort();
            }
            else {
                comm = new FlyAd2021Comm_TcpClient();
            }
            core = new FlyAd2021Core();

            comm.Addr = Addr;

            comm.RecMsg += (msg) => core.RecMsg(msg);

            core.SendMsgEvent += (s) =>
            {
                byte[] msg = core.GetSendMsg();
                comm.SendMsg(msg);
            };

            core.PushDataEvent += Core_PushDataEvent;
            core.PushRunResultEvent += Core_PushRunResultEvent;
            core.StartMeasure();
            comm.Start();

            core.PropertyChanged += Core_PropertyChanged;
        }

        private void Core_PushRunResultEvent(object sender, PushRunResultEventArgs e)
        {
            //TODO
            mSysTick.ToDateTime(e.SysTick);

            if (e.Status == DRIVE_MAN_STATUS.RUNNING) 
            {
                //
            }
        }

        /// <summary>
        /// 计数32位脉冲
        /// </summary>
        /// <param name="lastPostion">上一次脉冲</param>
        /// <param name="enc">低16位脉冲</param>
        /// <returns></returns>
        int CalPosition(int lastPostion, UInt16 enc)
        {
            //Position
            UInt16 lastEnc = (UInt16)(lastPostion & 0xffff);

            Int16 d = (Int16)(enc - lastEnc);

            lastPostion += d;

            return lastPostion;
        }


        private void Core_PushDataEvent(object sender, PushDataEventArgs e)
        {
            Now = mSysTick.ToDateTime(e.SysTick);
            AD = e.AD;
            if (e.ENC1 != null) 
            {
                Position = CalPosition(Position, (UInt16)e.ENC1);
            }
            if (e.ENC2 != null)
            {
                Position2 = CalPosition(Position2, (UInt16)e.ENC2);
            }
            if (e.In != null)
            {
                IStatus = (UInt16)e.In;
            }
            if (e.Out != null)
            {
                OStatus = (UInt16)e.Out;
            }


            //填充timegrid

        }

        private void Core_PropertyChanged(object sender, PropertyChangedEventArgs e)
        {
            if (e.PropertyName == nameof(core.IsConnected)) {

                if (core.IsConnected)
                {
                    //连接成功

                    if (IsReadParamFromDev)
                    {
                        //从设备读取参数

                        //获取当前全部状态
                        core.GetSysParam_Jog((asyncContext, retData) => JogVelocity = (UInt32)retData, null);
                        core.GetSysParam_Zero((asyncContext, retData) => PosOffset = (Int16)retData, null);
                        core.GetSysParam_MotorType((asyncContext, retData) => MotorType = (MOTORTYPE)retData, null);
                        core.GetSysParam_Ratio01((asyncContext, retData) => Ratio01 = (UInt16)retData, null);
                        core.GetSysParam_Ratio02((asyncContext, retData) => Ratio02 = (UInt16)retData, null);

                        core.GetRunParam_V((asyncContext, retData) => Velocity = (UInt32)retData, null);
                        core.GetRunParam_SV((asyncContext, retData) => SVelocity = (UInt32)retData, null);
                        core.GetRunParam_AccTime((asyncContext, retData) => ATime = (UInt32)retData, null);
                        core.GetRunParam_DecTime((asyncContext, retData) => DTime = (UInt32)retData, null);
                        core.GetRunParam_HSpd1((asyncContext, retData) => HVelocity1 = (UInt32)retData, null);
                        core.GetRunParam_HSpd2((asyncContext, retData) => HVelocity2 = (UInt32)retData, null);
                    }
                    else
                    {
                        //参数不保存在设备中， 保存在电脑。
                        //参数写入到设备。
                        NotifyPropertyChanged(nameof(MotorType));
                        NotifyPropertyChanged(nameof(Ratio01));
                        NotifyPropertyChanged(nameof(Ratio02));
                        NotifyPropertyChanged(nameof(PosOffset));
                        NotifyPropertyChanged(nameof(JogVelocity));
                    }

                    core.GetAccess((asyncContext, retData) =>
                    {
                        GetAccess_Reponse reponse = (GetAccess_Reponse)retData;
                        AreaStatus = reponse.status;
                        AreaRet = reponse.ret;

                        Array.Copy(reponse.code, Code, reponse.code.Length);
                        Code[6] = 0x06;
                        NotifyPropertyChanged(nameof(Code));

                        Surplus = reponse.surplus;

                        Array.Copy(reponse.access, Access, reponse.access.Length);
                        NotifyPropertyChanged(nameof(Access));

                    }, null);

                    core.GetRunResult((asyncContext, retData) =>
                    {
                        GetRunResult_Reponse reponse = (GetRunResult_Reponse)retData;
                        DriveStatus = reponse.result;
                        Marker = (Int32)reponse.serial;
                    }, null);

                    core.GetEncAll((asyncContext, retData) =>
                    {
                        var reponse = retData as GetEncAll_Reponse;
                        Position = reponse.pos1;
                        Position2 = reponse.pos2;
                    }, null);
                    core.GetIn((asyncContext, retData) =>
                    {
                        var reponse = (UInt16)retData;
                        IStatus = reponse;
                    }, null);
                    core.GetOut((asyncContext, retData) =>
                    {
                        var reponse = (UInt16)retData;
                        OStatus = reponse;
                    }, null);


                }
                else 
                {
                    //连接断开，复位 脉冲计数。。。,时间计数
                    //TODO
                }
            }
        }


        /// <summary>
        /// Out0 = 0, SetOutput(0x0001,0x0000);
        /// Out0 = 1, Out1 = 0, SetOutput(0x0003,0x0001);
        /// </summary>
        /// <param name="mask"></param>
        /// <param name="enable"></param>
        public void SetOutput(UInt16 mask, UInt16 enable)
        {
            core.SetOutPorts(mask, enable, null, null);
        }

        /// <summary>
        /// 读取参数, 全部！！！！
        /// </summary>
        public void UpdateParam()
        {
            core.GetSysParam_Jog((asyncContext, retData) => JogVelocity = (UInt32)retData, null);
            core.GetSysParam_Zero((asyncContext, retData) => PosOffset = (Int16)retData, null);
            core.GetSysParam_MotorType((asyncContext, retData) => MotorType = (MOTORTYPE)retData, null);
            core.GetSysParam_Ratio01((asyncContext, retData) => Ratio01 = (UInt16)retData, null);
            core.GetSysParam_Ratio02((asyncContext, retData) => Ratio02 = (UInt16)retData, null);

            core.GetAccess((asyncContext, retData) =>
            {
                GetAccess_Reponse reponse = (GetAccess_Reponse)retData;
                AreaStatus = reponse.status;
                AreaRet = reponse.ret;

                Array.Copy(reponse.code, Code, reponse.code.Length);
                Code[6] = 0x06;
                NotifyPropertyChanged(nameof(Code));

                Surplus = reponse.surplus;

                Array.Copy(reponse.access, Access, reponse.access.Length);
                NotifyPropertyChanged(nameof(Access));

            }, null);

            core.GetRunParam_V((asyncContext, retData) => Velocity= (UInt32)retData, null);
            core.GetRunParam_SV((asyncContext, retData) => SVelocity = (UInt32)retData, null);
            core.GetRunParam_AccTime((asyncContext, retData) => ATime = (UInt32)retData, null);
            core.GetRunParam_DecTime((asyncContext, retData) => DTime = (UInt32)retData, null);
            core.GetRunParam_HSpd1((asyncContext, retData) => HVelocity1 = (UInt32)retData, null);
            core.GetRunParam_HSpd2((asyncContext, retData) => HVelocity2 = (UInt32)retData, null);

            core.GetRunResult((asyncContext, retData)=>{
                GetRunResult_Reponse reponse = (GetRunResult_Reponse)retData;
                DriveStatus = reponse.result;
                Marker = (Int32)reponse.serial;
            },null);
        }

        #region 运动参数
        /// <summary>
        /// 设置 运动参数
        /// </summary>
        /// <param name="posOfGrid">脉冲/grid</param>
        /// <param name="motortype">电机类型</param>
        /// <param name="ratio01">脉冲比例 分母(电机脉冲)</param>
        /// <param name="ratio02">脉冲比例 分子(编码器脉冲)</param>
        public void SetSysParam(UInt16 posOfGrid, MOTORTYPE motortype, UInt16 ratio01, UInt16 ratio02) 
        {
            //TODO
            //posOfGrid 没有了，需要自己实现

            core.SetSysParam_MotorType(motortype, null, null);
            core.SetSysParam_Ratio01(ratio01, null, null);
            core.SetSysParam_Ratio02(ratio02, null, null);
        }
        /// <summary>
        /// 电机类型
        /// </summary>
        public MOTORTYPE MotorType { get; set; } = MOTORTYPE.SERVO;
        /// <summary>
        /// 脉冲/grid, 最大1000个grid
        /// </summary>
        private UInt16 posofgrid = 10;
        /// <summary>
        /// 脉冲/grid, 最大1000个grid
        /// </summary>
        public UInt16 PosOfGrid
        {
            get { return posofgrid; }
            set
            {
                if (value < 1)
                    value = 1;

                if (posofgrid != value)
                    posofgrid = value;

            }
        }
        /// <summary>
        /// 脉冲比例 分母(电机脉冲)
        /// </summary>
        public UInt16 Ratio01 { get; set; } = 4;
        /// <summary>
        /// 脉冲比例 分子(编码器脉冲)
        /// </summary>
        public UInt16 Ratio02 { get; set; } = 1;

        #endregion

        /// <summary>
        /// 脉冲平移(0位处脉冲值)
        /// </summary>
        public Int16 PosOffset { set; get; }
        /// <summary>
        /// 手动速度 pps
        /// </summary>
        public UInt32 JogVelocity { set; get; } = 1000;

        #region 速度参数
        /// <summary>
        /// 设置速度参数
        /// </summary>
        /// <param name="velocity">runto 速度 pps</param>
        /// <param name="sv">开始速度 pps</param>
        /// <param name="atime">加速时间 ms</param>
        /// <param name="dtime">减速时间 ms</param>
        /// <param name="hspeed1">归0 第1段速 pps</param>
        /// <param name="hspeed2">归0 第2段速 pps</param>
        public void SetPosParam(UInt32 velocity, UInt32 sv, UInt32 atime, UInt32 dtime, UInt32 hspeed1, UInt32 hspeed2) 
        {
            core.SetRunParam_V(velocity, null, null);
            core.SetRunParam_SV(sv, null, null);
            core.SetRunParam_AccTime(atime, null, null);
            core.SetRunParam_DecTime(dtime, null, null);
            core.SetRunParam_HSpd1(hspeed1, null, null);
            core.SetRunParam_HSpd2(hspeed2, null, null);
        }
        /// <summary>
        /// 设置速度
        /// </summary>
        /// <param name="velocity"></param>
        public void SetVelocity(UInt32 velocity) 
        {
            core.SetRunParam_V(velocity, null, null);
        }
        /// <summary>
        /// runto 速度 pps
        /// </summary>
        public UInt32 Velocity { get; private set; }
        /// <summary>
        /// 开始速度 pps
        /// </summary>
        public UInt32 SVelocity { get; private set; }
        /// <summary>
        /// 加速时间 ms
        /// </summary>
        public UInt32 ATime { get; private set; }
        /// <summary>
        /// 减速时间 ms
        /// </summary>
        public UInt32 DTime { get; private set; }
        /// <summary>
        /// 归0 第1段速 pps
        /// </summary>
        public UInt32 HVelocity1 { get; private set; }
        /// <summary>
        /// 归0 第2段速 pps
        /// </summary>
        public UInt32 HVelocity2 { get; private set; }

        #endregion

        #region 密码
        /// <summary>
        /// 使用码区状态
        /// </summary>
        public AREA_STATUS AreaStatus { get; private set; }
        /// <summary>
        /// 使用码区出错码
        /// </summary>
        public AREA_ERR AreaRet { get; private set; }

        /// <summary>
        /// 序列码byte[6],  AD盒只需要 6个byte，但flysecute 软件需要 7byte， code[6]=0x06
        /// </summary>
        public byte[] Code { get; private set; } = new byte[7];
        /// <summary>
        /// 系统剩余时间
        /// </summary>
        public int Surplus { get; private set; }
        /// <summary>
        /// 系统授权码access[8]
        /// </summary>
        public byte[] Access { get; private set; } = new byte[8];
        /// <summary>
        /// 输入系统授权码
        /// </summary>
        /// <param name="access">系统授权码access[8]</param>
        public void SetAccess(byte[] access) 
        {
            core.CheckAccessCode(access, (asyncContext, retData) =>
            {
                AreaRet = (AREA_ERR)retData;
            },null);
        }
        /// <summary>
        /// 初始化系统信息区
        /// </summary>
        public void InitArea() 
        {
            core.InitArea(null, null);
        }

        #endregion


        /// <summary>
        /// timegrid 事件; 版本1: 1.28s 一次; 版本2: 0.2s一次， 
        /// </summary>
        public event TimeGridEventHandler TimeGridEvent;

        /// <summary>
        /// runto 中，每经过1个grid 推送数据
        /// </summary>
        public event MiniGridEventHandler MiniGridEvent;

        /// <summary>
        /// runto 的 grid事件
        /// </summary>
        public event MiniGridEventHandler GridEvent;

        /// <summary>
        /// 输入状态改变事件
        /// </summary>
        public event IStatusChangedEventHandler IStatusChangedEvent;

        /// <summary>
        /// 指令状态改变事件
        /// </summary>
        public event DriveStatusChangedEventHandler DriveStatusChangedEvent;

        /// <summary>
        /// 脉冲1改变事件
        /// </summary>
        public event PositionChangedEventHandler PositionChangedEvent;

        /// <summary>
        /// 脉冲2改变事件
        /// </summary>
        public event PositionChangedEventHandler Position2ChangedEvent;


        #region 运行动作
        /// <summary>
        /// 运行至指定位置
        /// </summary>
        /// <param name="to">脉冲</param>
        public void Runto(int to) 
        {
            //TODO
            core.RunTo(to, 123,null, null);
            this.DriveOrder = DRIVE_MAN_ORDER.RUNTO;
            this.DriveStatus = DRIVE_MAN_STATUS.RUNNING;
        }
        /// <summary>
        /// 归原点
        /// </summary>
        public void Origin() 
        {
            //TODO
            core.Org(123,null,null);
        }
        /// <summary>
        /// 停止
        /// </summary>
        public void Stop() 
        {
            core.Stop(null, null);
        }
        /// <summary>
        /// 反向，直到撞了限位
        /// </summary>
        public void Backward() 
        {
            core.Backw(123, null, null);
        }
        /// <summary>
        /// 正行，直到撞了限位
        /// </summary>
        public void Forward() 
        {
            core.Forw(123, null, null);
        }

        /// <summary>
        /// 当前运行指令
        /// </summary>
        public DRIVE_MAN_ORDER DriveOrder { get; private set; }
        /// <summary>
        /// 之前运行动作的状态
        /// </summary>
        public DRIVE_MAN_STATUS DriveStatus { get; private set; }
        #endregion


        #region 同步
        /// <summary>
        /// 同步用标示
        /// </summary>
        public Int32 Marker { get; private set; }

        /// <summary>
        /// 逻辑横向脉冲偏移  Pos1 + Pos1LCShift = Pos1LC
        /// </summary>
        public int Pos1LCShift { get; set; }
        /// <summary>
        /// 逻辑纵向脉冲偏移  Pos2 + Pos2LCShift = Pos2LC
        /// </summary>
        public int Pos2LCShift { get; set; }


        #region 同步控制 同步状态转换规则

        /// <summary>
        /// 进入同步状态
        /// </summary>
        public void SyncBegin() 
        {
            core.SyncRun_Start(null, null);
        }
        /// <summary>
        /// 退出同步状态
        /// </summary>
        public void SyncEnd()
        {
            core.SyncRun_Stop(null, null);
        }
        /// <summary>
        /// 清空同步指令
        /// </summary>
        public void SyncClear()
        {
            core.SyncRun_Clear(null, null);
        }

        #endregion
        #region 同步扫描 脚本指令

        /// <summary>
        /// 同步扫描至;
        /// D+0xE0+开始主轴位置+结束主轴位置+结束横向脉冲位置（逻辑位置）+脉冲开关（1B）+命令识标号（4B）
        /// </summary>
        /// <param name="pos2lc_begin"></param>
        /// <param name="pos2lc_end"></param>
        /// <param name="pos1lc"></param>
        /// <param name="hasDataGrid"></param>
        /// <param name="marker"></param>
        public void SyncRunAtLC(int pos2lc_begin, int pos2lc_end, int pos1lc, bool hasDataGrid, Int32 marker) 
        {
            //TODO, 怎样解决 grid 数据？
            core.SyncRun_RunAtLC(pos2lc_begin, pos2lc_end, pos1lc, (UInt32)marker, null, null);
        }


        /// <summary>
        /// 位于队列头时运行，归零；
        /// D+0xE1+命令识标号（4B）
        /// </summary>
        /// <param name="marker"></param>
        public void SyncOrigin(Int32 marker) 
        {
            core.SyncRun_Origin((UInt32)marker, null, null);
        }


        /// <summary>
        /// 位于队列头时运行，以速度运行至物理位置；
        /// D+0xE2+横向脉冲位置（4B：int32，物理位置）+速度（4B：int32）+脉冲开关（1B）+命令识标号（4B）
        /// </summary>
        /// <param name="pos1"></param>
        /// <param name="velocity"></param>
        /// <param name="hasDataGrid"></param>
        /// <param name="marker"></param>
        public void SyncRunTo(int pos1, UInt32 velocity, bool hasDataGrid, Int32 marker) 
        {
            //TODO, 怎样解决 grid 数据？
            core.SyncRun_RunTo(pos1, velocity, (UInt32)marker, null, null);
        }


        /// <summary>
        /// 位于队列头时运行，以速度运行至逻辑位置；
        /// D+0xE3+横向脉冲位置（4B：int32，逻辑位置）+速度（4B：int32）+脉冲开关（1B）+命令识标号（4B）
        /// </summary>
        /// <param name="pos1lc"></param>
        /// <param name="velocity"></param>
        /// <param name="hasDataGrid"></param>
        /// <param name="marker"></param>
        public void SyncRunToLC(int pos1lc, UInt32 velocity, bool hasDataGrid, Int32 marker) 
        {
            core.SyncRun_RunToLC(pos1lc, velocity, (UInt32)marker, null, null);
        }


        /// <summary>
        /// 等待，ms；
        /// D+0xE4+毫秒数（4B：int32）+命令识标号（4B）
        /// </summary>
        /// <param name="ms"></param>
        /// <param name="marker"></param>
        public void SyncWait(int ms, Int32 marker) 
        {
            //没有实现！！！！
        }

        #endregion
        #endregion
    }

    /// <summary>
    ///  systick 转 DateTime
    ///  不断把 从flyad7 获取的 systick 放入。  以每次systick 的差 加上 上一次 dt_1st, 得到时间。
    ///  如： 当前 获得 systick1, 之前是 systick0,   所以 dt = dt_1st+ (systick1-systick0);  dt_1st=dt;
    ///  
    /// 第1个 dt_1st, 是 从 DateTime.Now. 获取的。
    /// 
    /// 不会管 当前DateTime.Now 与 dt 差多少。 所以如果 flyad7 时间比较慢，可能会出现 3天后， DateTime.Now 与 dt 差了 1个小时。
    /// </summary>
    class TSysTick
    {
        byte systick_1st = 0;
        DateTime dt_1st = DateTime.MinValue;

        /// <summary>
        /// 
        /// </summary>
        public event Action BeReseted;
        public void Reset()
        {
            dt_1st = DateTime.MinValue;
        }
        /// <summary>
        /// 输入从 flyad7 获取的 timemark, 转换为 本地 DateTime
        /// </summary>
        /// <param name="systick">只有低7位有效</param>
        /// <returns></returns>
        public DateTime ToDateTime(byte systick)
        {
            if (dt_1st == DateTime.MinValue)
            {
                dt_1st = DateTime.Now;
                systick_1st = systick;
                return dt_1st;
            }

            int ms = SystickSub(systick, systick_1st);
            if ((ms > 30))//记录与新的timemark 相差30ms，掉包很严重
            {
                //被复位了
                BeReseted?.Invoke();
            }

            DateTime dt = dt_1st.AddMilliseconds(ms);

            systick_1st = systick;
            dt_1st = dt;
            return dt;
        }

        /// <summary>
        /// 7位 systick1-systick2
        /// </summary>
        /// <param name="systick1"></param>
        /// <param name="systick2"></param>
        /// <returns></returns>
        int SystickSub(byte systick1, byte systick2) 
        {
            if (systick1 >= systick2)
                return systick1 - systick2;
            else {
                return systick1 + 0x80 - systick2;
            }
        }
    }



}
